#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_recipe
----------------------------------

Tests for `inicheck.Tools` module.
"""

import unittest
from inicheck.tools import *
from collections import OrderedDict

class TestTools(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        base = os.path.dirname(__file__)
        self.ucfg = get_user_config(os.path.join(base,"test_configs/full_config.ini"),
                                                            modules="inicheck")

    def test_get_checkers(self):
        """
        Tests the get_checkers func in tools
        """
        checkers = get_checkers().keys()

        valids = ['bool', 'criticaldirectory', 'criticalfilename', 'datetime',
                  'datetimeorderedpair', 'directory', 'filename', 'float',
                  'int', 'string', 'url']

        for v in valids:
            assert v in checkers

        valids = ["type", "generic", "path"]
        checkers = get_checkers(ignore=[]).keys()
        for v in valids:
                assert v in checkers

    def test_check_config(self):
        """
        Tests the check_config func in tools
        """
        ucfg = self.ucfg

        warnings, errors = check_config(ucfg)

        assert len(errors) == 12

    def test_cast_all_variables(self):
        """
        Tests the cast_all_variables func in tools
        """
        ucfg = self.ucfg

        ucfg.cfg['topo']['test_start'] = "10-1-2019"
        ucfg.cfg['air_temp']['dk_ncores'] = "1.0"
        ucfg.cfg['air_temp']['detrend'] = "true"

        ucfg = cast_all_variables(ucfg, ucfg.mcfg)
        results = ['timestamp','int','bool','float','list','str']

        tests = [ucfg.cfg['time']['start_date'],
                  ucfg.cfg['air_temp']['dk_ncores'],
                  ucfg.cfg['air_temp']['detrend'],
                  ucfg.cfg['wind']['reduction_factor'],
                  ucfg.cfg['output']['variables'],
                  ucfg.cfg['air_temp']['distribution']]

        for i,v in enumerate(tests):
            assert results[i] in type(v).__name__.lower()


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



from . iniparse import parse_sections, parse_items, parse_changes
from pandas import to_datetime
from . entries import ConfigEntry
from . utilities import mk_lst
import importlib
from os.path import join as pjoin
from os.path import abspath
from collections import OrderedDict

class ChangeLog(object):

    def __init__(self, paths=None, mcfg=None, **kwargs):

        # Paths to changelogs
        self.paths = []

        # We got a direct path
        if paths != None:
            paths = mk_lst(paths)
            self.paths += paths

            self.changes = self.join_changes(self.paths)

        self.check_log_validity(mcfg)

    def join_changes(self, paths):
        """
        Joins multiple changelogs together

        Args:
            paths: List of paths containing changelogs syntax
        Returns:
            changes: lists of line item changes provided in each line as a
                     list of [old, new]
        """

        changes = []
        for p in paths:
            changes += self.read_change_log(p)

        return changes

    def read_change_log(self, path):
        """
        Opens the file and reads in sections and the items.

        Args:
            path: Path to a changlog file
        """
        with open(path, encoding='utf-8') as f:
            lines = f.readlines()
            f.close()

        sections = parse_sections(lines)
        raw_changes = sections['changes'].copy()
        del sections['changes']

        # Parse meta data the same way as everything
        sec_and_items = parse_items(sections)
        self.info = sec_and_items['meta']['info']
        self.date = to_datetime(sec_and_items["meta"]['date'])

        # Parse the change list
        changes = parse_changes(raw_changes)

        return changes

    def check_log_validity(self, mcfg):
        """
        Checks the current master config that the items we're moving to
        are valid. Also confirms that removals are aligned with the master.
        """
        action_kw = ["any",'removed']
        cfe = ConfigEntry()
        invalids = []

        for zz,c in enumerate(self.changes):
            # Check the new assignments match the names of current master items
            current = c[1]
            valids = []

            # KW removed or ANY is valid for sections
            if current[0] in action_kw or current[0] in mcfg.cfg.keys():
                valids.append(True)

            # Section is valid, lets check items
            if current[0] != "removed":

                if len(valids) >= 1:

                    # Non-any item and any section provided
                    if current[0] == "any" and current[1] not in action_kw:
                        for s, items in mcfg.cfg.items():
                            if current[1] in items.keys():
                                valids.append(True)
                                break

                    # Valid item check whe valid section provided
                    elif (current[1] in action_kw or
                          current[1] in mcfg.cfg[current[0]].keys()):
                        valids.append(True)

                # Check for valid properties
                if len(valids) >= 2:
                        if current[2] in ["any", "default"]:
                            valids.append(True)

                # TODO Actually check values, ignoring for now.
                if len(valids) == 3:
                    valids.append(True)

                # Final check
                if len(valids) != len(current):
                    invalids.append(c)

        # Form a coherent message about incorrect changlog stuff
        if invalids:
            msg = ("Changelog states a change that doesn't match the core config."
                  " For a change to be valid the new changes must be in the"
                  " Master Config file. Mismatches are:")

            for n in invalids:
                msg+="\n * "
                msg+="/".join(n[0])
                msg+=" -> "
                msg+="/".join(n[1])

            raise ValueError(msg)

    def get_active_changes(self, ucfg):
        """
        Goes through the users config looking for matching old configurations,
        then makes recommendations.

        Args:
            ucfg: UserConfig Object
        """
        required_changes = []
        potential_changes = []

        cfe = ConfigEntry()

        cfg = ucfg.cfg

        for change in self.changes:

            # Go through the sections
            for s in cfg.keys():

                # Go through the items
                for i in cfg[s].keys():

                    # Assign original changes to any's
                    assumed = change[0].copy()
                    new = change[1].copy()

                    # Swap out anys with section names
                    if change[0][0] == "any":
                        assumed[0] = s
                        if change[1][0] != "removed":
                            new[0] = s

                    # swap out anys with item names
                    if change[0][1] == "any":
                        assumed[1] = i
                        if change[1][0] != "removed":
                            new[1] = i

                    # If we have a match from the changelog and the ucfg
                    if assumed[0] == s and assumed[1] == i:
                        if "removed" in new:
                            required_changes.append([assumed, "removed"])

                        # Make sure "any" doesn't disagree with master
                        elif new[0] in ucfg.mcfg.cfg.keys():
                            if new[1] in ucfg.mcfg.cfg[new[0]].keys():

                                # Check for an old default match and suggest a change
                                if assumed[2] == "default":

                                    value = str(mk_lst(cfg[s][i], unlst=True))
                                    if value == assumed[3]:
                                        potential_changes.append([assumed, new])

                                else:
                                    required_changes.append([assumed, new])

        return potential_changes, required_changes

    def apply_changes(self, ucfg, potentials, changes):
        """
        Iterate through the list of detectd applicable changes retrieved
        from get_actived_changes to produce a new config file with the
        correct changes.

        Args:
            ucfg: UserConifg Object
            changes: list lists length 4 (section item property value)
                    representing detected changes

        Returns:
            cfg: Config dictionary
        """

        cfg = ucfg.cfg.copy()

        # REVIEW Right now only defaults can be changed so we assume that here
        for p in potentials:
            s_o = p[0][0]
            i_o = p[0][1]

            # assign the new defaults
            cfg[s_o][i_o] = p[1][3]

        for c in changes:
            removal = False
            s_o = c[0][0]
            i_o = c[0][1]

            # Avoid adding removed items
            if c[1] == "removed":
                removal = True

            else:
                s_n = c[1][0]
                i_n = c[1][1]

            #print(s_o, i_o, s_n, i_n)

            # Confirm new section in the config
            if not removal:
                if s_n not in ucfg.cfg.keys():
                    cfg[s_n] = OrderedDict()

            # Its been deprecated
            if removal:
                del cfg[s_o][i_o]

            # Item or section is a name transfer.
            elif s_o in ucfg.cfg.keys():
                if i_o in cfg[s_o].keys():

                    if s_n != "removed":
                        cfg[s_n][i_n] = cfg[s_o][i_o]

                    del cfg[s_o][i_o]

                # look to remove a whole section
                if len(cfg[s_o].keys()) == 0 and s_o not in ucfg.mcfg.cfg.keys():
                    del(cfg[s_o])

        return cfg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_config
----------------------------------

Tests for `inicheck.config` module.
"""

import unittest
import os
from inicheck.config import *
from inicheck.tools import cast_all_variables, get_checkers
import datetime

def compare_config(generated_config, truth_config, master=False, mcfg_attr='default'):
    """
    Compares config objects for quick tests:

    Args:
        generated_config: the config file the inicheck system produced automatically
        truth_config: a dictionary made up for checking against
        master: boolean for identifying if were checking master config
        mcfg_attr: attribute we want to check when using a master config
    Returns:
        result: boolean describing if cfg are the same.
    """

    # First confirm all keys are available
    for s in truth_config.keys():
        if s  not in generated_config:
            print("\nERROR: test data contains section: {} that was not"
                  " contained by the generated config.".format(s))
            return False

        for i,v in truth_config[s].items():

            # Then confirm the item is in the section in the generated config
            if i not in generated_config[s].keys():
                print("\nERROR: test data contains item: {} in section: {} that was not"
                      " contained by the generated config.".format(i,s))
                return False

            # Check the value
            else:
                # if it is a master config object then use it differently.
                if master:
                    gv = getattr(generated_config[s][i], mcfg_attr)
                    emsg = "\nERROR: config.{} = {} not {}.".format(
                                                    mcfg_attr,
                                                    gv,
                                                    v)

                # Normal config
                else:
                    gv = generated_config[s][i]
                    emsg ="\nERROR: config[{}][{}] = {} not {}.".format(s,i,gv,
                                                                              v)
                # Confirm values are the same
                if gv != v:
                        print(emsg)
                        return False

    return True


class TestUserConfig(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        """
        """
        fname = os.path.abspath(os.path.dirname(__file__)+'/test_configs/full_config.ini')
        mcfg = MasterConfig(modules='inicheck')
        self.ucfg = UserConfig(fname, mcfg=mcfg)

    def test_ucfg_init(self):
        """
        Simply opens and looks at the base config BEFORE all the recipes and
        stuff has been applied
        """

        # Assert important attributes
        for a in ['mcfg','cfg','raw_cfg','recipes']:
            assert(hasattr(self.ucfg, a))


    def test_apply_recipes(self):
        """
        Tests that the correct recipes were identified to be used for
        interpretation
        """

        truth = {'topo':
                        {'filename':['None'],
                        'type':'netcdf'},
                 'cloud_factor':
                        {'distribution': 'idw'}
                }

        self.ucfg.apply_recipes()
        valid_recipes = ['topo_basic_recipe', 'time_recipe', 'air_temp_recipe',
                         'vp_recipe', 'wind_recipe', 'precip_recipe',
                         'non_winstral_recipe', 'cloud_factor_recipe',
                         'albedo_recipe', 'date_decay_method_recipe',
                         'solar_recipe', 'thermal_recipe', 'soil_recipe',
                         'output_recipe', 'system_recipe', 'csv_recipe',
                         'remove_wind_ninja_recipe', 'non_grid_local_recipe',
                         'dk_recipe', 'idw_recipe']
        for v in valid_recipes:
            assert v in [r.name for r in self.ucfg.recipes]


class TestMasterConfig(unittest.TestCase):
    def setUp(self):
        """
        Stage our truthing data here
        """
        self.truth_defaults = {'topo':
                                      {'type': 'netcdf',
                                       'filename': ['./common_data/topo/topo.nc']},
                               'air_temp':
                                       {'distribution': 'idw',
                                        'detrend': 'true',
                                        'dk_ncores': '1'}}

    def test_grabbing_mcfg(self):
        """
        Builds a master config from the module and paths, check it.
        """
        # Build a master config file using multiple files
        try:
            mcfg = MasterConfig(modules = 'inicheck')
            assert True
        except:
            assert False

        base = os.path.dirname(__file__)
        master = os.path.join(base,"./test_configs/CoreConfig.ini")
        recipes = os.path.join(base,"./test_configs/recipes.ini")

        try:
            mcfg = MasterConfig(path = [master,recipes])
            assert True
        except:
            assert False

    def test_add_files(self):
        """
        Builds a master config from the files, check it.
        """
        # Build a master config file using multiple files
        base = os.path.dirname(__file__)
        master = os.path.join(base,"test_configs/CoreConfig.ini")
        recipes = os.path.join(base,"test_configs/recipes.ini")

        mcfg = MasterConfig(path=master)
        mcfg.cfg = mcfg.add_files([master,recipes])

        valid_sections = ['topo','csv','air_temp']
        for v in valid_sections:
            assert v in mcfg.cfg.keys()

        assert 'topo_basic_recipe' in [r.name for r in mcfg.recipes]


    def test_check_types(self):
        """
        Checks to make sure we throw the correct error when an unknown data
        type is requested
        """

        # Call out a BS entry type to raise the error
        checkers = get_checkers()
        invalids = ['str','filepath']
        valids = ['bool', 'criticaldirectory', 'criticalfilename',
                  'discretionarycriticalfilename', 'datetime',
                  'datetimeorderedpair', 'directory', 'filename', 'float',
                  'int', 'string', 'url']

        for z, values in enumerate([invalids,valids]):
            for kw in values:
                line = ["type = {}".format(kw), "description = test"]
                cfg = {"section":{"test": ConfigEntry(name='test',
                                                      parseable_line=line)}}

                if z == 0:
                    self.assertRaises(ValueError, check_types, cfg, checkers)

                else:
                    assert check_types(cfg, checkers)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



import os
from .utilities import  remove_chars, remove_comment
from collections import OrderedDict


def read_config(fname):
    """
    Opens and reads in the config file in its most raw form.
    Creates a dictionary of dictionaries that contain the string result

    Args:
        fname: Real path to the config file to be opened
    Returns:
        config: dict of dicts containing the info in a config file
    """
    with open(fname, encoding = 'utf-8') as f:
        lines = f.readlines()
        f.close()

    sections = parse_sections(lines)
    sec_and_items = parse_items(sections)
    config = parse_values(sec_and_items)

    return config


def parse_entry(info, item=None, valid_names=None):
    """
    Parse the values found in the master config where the entries are a little
    more complicated. Specifically this deals with syntax issues and whether
    a entry option is valid or not.

    This should be used only after knowning the section and item.

    Args:
        info: Potential line to parse containing masater config options
        item: Item name were parsing. Only for reporting errors purposes
        valid_names: valid property names to be parsing.

    Returns:
        properties: dictionary containing the properties as keys
    """

    properties = OrderedDict()
    if type(info) != list:
        info = [info]

    last_three = []

    for s in info:

        if '=' in s:
            a = s.split('=')

        else:
            raise ValueError('\n\nMaster Config file missing an equals sign in'
                            ' entry or missing a comma right before the item '
                            '"{0}" in the entry:\n"{1}"\nIssue generated from:'
                            ' \n>> "{2}"'.format(item, info, s))

        name = (a[0].lower()).strip()

        # Check for constraints on potential names of entries
        if valid_names != None and name not in valid_names or len(a) > 2:
            raise ValueError("\nInvalid property set in the master config file,"
                            " available options are: \n * {0}\n"
                            "\nIf this is supposed to be a recipe, you may have"
                            " forgotten to add keyword 'recipe' to the section"
                            " name.\nIssue generated from: \n>> '{1}'"
                            "".format("\n * ".join(valid_names), s))

        value = a[1].strip()

        result = []
        value = value.replace('\n'," ")
        value = value.replace('\t',"")

        # Is there a list of values provided?
        if '[' in value:
            if ']' not in value:
                raise ValueError('Missing bracket or commas used in a list'
                                 ' instead of spaces in config file under'
                                 ' {0} in the entry:\n"{1}"'.format(name, info))
            else:
                value = (''.join(c for c in value if c not in '[]'))
                value = value.split(' ')

        properties[name] = value
    return properties


def parse_sections(lines):
    """
    Returns a dictionary containing all the sections as keys with a single
    string of the contents after the section

    Args:

        lines: a list of string lines containing all the raw info of a cfg file

    Returns:
        sections: dictionary containing keys that are the section names and
                  values that are strings of the contents between sections

    """

    result = OrderedDict()
    section = None
    i = 0

    for i in range(len(lines)):
        line = remove_chars(lines[i],'\t')
        # Comment checking
        line = remove_comment(line)

        # Watch out for extraneous chars at the beginning an end
        line = line.strip()

        # Check for empty line first
        if line and line not in os.linesep:
            # Look for section
            if line.startswith('['):
                # Look for open brackets
                if ']' in line:

                    section = (remove_chars(line,'[]')).lower().strip()
                    if section in result.keys():
                        raise ValueError("Section name {} already used in "
                                         "config, consider renaming it to "
                                         "something unique.".format(section))
                    result[section] = []

            else:
                # This protects from funky syntax in a config file and alerts the user
                if section == None:
                    raise Exception("Non-section like syntax before any "
                                    "sections were identified at line {0} in "
                                    "config file. Please use bracketed sections"
                                    " or use # or ; to write comments."
                                    "".format(i))
                else:
                    result[section].append(line)

    return result


def parse_items(parsed_sections_dict, mcfg=None):
    """
    Takes the output from parse_sections and parses the items in each
    section

    Args:
        parsed_sections_dict: dictionary containing keys as the sections
                              and values as a continuous string of the
    Returns:
        result: dictionary of dictionaries containing sections,items, and
                their values
    """
    result = OrderedDict()

    for k,v in parsed_sections_dict.items():
        item = None
        result[k] = OrderedDict()
        for i,val in enumerate(v):
            val = val.lstrip()
            # Look for item notation

            if ':' in val:
                # Only split on the first colon to avoid collisions with datetime
                parse_loc = val.index(':')
                parseable = [val[0:parse_loc],val[parse_loc + 1:]]
                item = parseable[0].lower().strip()

                result[k][item] = ''

                # Check for a value right after the item name
                potential_value = (parseable[-1].replace('\n', ' ')).lstrip()

                # Avoid stashing properties in line with the item
                if '=' not in potential_value:
                    result[k][item] = potential_value

                # Property value provided in line with item
                else:
                    result[k][item] += " " + potential_value

            # User added line returns likely for readability
            else:
                result[k][item] += " " + val

        # Perform a final cleanup
        if item != None:
            if ',' in result[k][item]:
                final = [e.strip() for e in result[k][item].split(',')]
                result[k][item] = ", ".join(final)

            result[k][item] = result[k][item].strip()

    return result


def parse_values(parsed_items):
        """
        Takes the output from parse_items and parses any values or
        properties provided in each item placing the strings into a list.
        If no properties are defined then it will clean up values provided
        and put them in a list of len one.

        Args:
            parsed_sections_dict: dict of dicts containing joined strings
                                  found under each item
        Returns:
            result: dictionary of dictionaries containing sections, items,
                    and the values provided as a list
        """
        result = OrderedDict()

        for section in parsed_items.keys():
            result[section] = OrderedDict()

            for item, val in parsed_items[section].items():
                value = val.strip()

                # list was provided
                if ',' in val:
                    final = [v.strip() for v in value.split(',') if v != '']

                # For use later always make it a list
                else:
                    final = [value]

                # Assign it finally and replace none
                result[section][item] = final

        return result


def parse_changes(change_list):
    """
    Takes in a section and item parsed dictionary and parses on > forming a
    list. It works as the following:

    To move an item to another section or item name
    section/item -> new_section/new_item

    section/item -> Removed

    Args:
        change_list: List of modifcations that occured to the config
    Returns:
        result: a list of a list length 2 containing of list length 4
                representing section item property value which are all any
                by default.
    """
    results = []

    for i,s in enumerate(change_list):
        if "->" in s:
            changes = [c.lower().strip() for c in s.split("->")]
        else:
            raise ValueError("Invalid changelog syntax at line "
                             " {}. A changelog entry must follow"
                             " <section>/<item>/<property>/<value> ->"
                             " <section>/<item>/<property>/<value>  or REMOVED"
                             "".format(i))

        # build out an any list and populate it
        final_changes = []

        for ii,c in enumerate(changes):
            mods = ["any" for i in range(4)]
            mod_lst = []
            if "/" in c:
                mod_lst = c.split("/")

            # Catch singular sections referenced
            else:
                mod_lst.append(c)

            mods[0:len(mod_lst)] = mod_lst
            final_changes.append(mods)

        results.append(final_changes)

    return results



import os
import sys
import inspect
from . config import UserConfig, MasterConfig, check_types
from . utilities import mk_lst, get_inicheck_cmd
import inicheck.checkers
from . changes import ChangeLog


def get_checkers(module='inicheck.checkers', keywords="check",
                                             ignore=["type","generic","path"]):
    """
    Args:
        module: The module to search for the classes
        keyword: Keywords to look for in class names
        ignore: Post parsed keywords to ignore if found in a class name

    Returns:
        dictionary: Dict of the classes available for checking config entries
    """
    keywords = mk_lst(keywords)

    funcs = inspect.getmembers(sys.modules[module], inspect.isclass)
    func_dict = {}

    for name, fn in funcs:
        checker_found = False
        k = name.lower()

        # Remove any keywords
        for w in keywords:
            z = w.lower()

            if z in k:
                k = k.replace(z, '')

                if k not in ignore:
                    checker_found = True
                    break

        if checker_found:
            func_dict[k] = fn

    return func_dict

def get_merged_checkers(ucfg):
    """
    Retrieve the dictionary of checker classes by grabbing all in inicheck and
    any prescribed in the master config through another module

    Args:
        ucfg: User Config object containing modules
    Returns:
        dictionary: all_checks - dictionary of all the checkers from inicheck
                    and any modules assigned to the master config.
    """

    # Grab all the original
    all_checks = get_checkers()

    # Add any checker modules if provided
    if ucfg.mcfg.checker_modules:
        for c in ucfg.mcfg.checker_modules:
            new_checks = get_checkers(module=c)
            all_checks.update(new_checks)

    return all_checks

def check_config(config_obj):
    """
    Looks at the users provided config file and checks it to a master
    config file looking at correctness and missing info.

    Args:
        config_obj - UserConfig object produced by
                     :class:`~inicheck.config.UserConfig`
    Returns:
        tuple:
        - **warnings** - Returns a list of string messages that are
                         consider non-critical issues with config file.
        - **errors** - Returns a list of string messages that are
                       consider critical issues with the config file.
    """

    msg = "{: <20} {: <30} {: <60}"
    errors = []
    warnings = []

    mcfg = config_obj.mcfg.cfg
    cfg = config_obj.cfg

    # Grab a dictionary of all the checkers
    all_checks = get_merged_checkers(config_obj)

    # Compare user config file to our master config
    for s, configured in cfg.items():

        # Section does not exists in master config
        if s not in mcfg.keys():
            err = "Not a valid section."
            errors.append(msg.format(s, " ", err))

        else:

            for i in configured.keys():
                # lower case item
                li = i.lower()

                # Item does not exist in the Master Config
                if li not in mcfg[s].keys():
                    wrn = "Not a registered option."
                    warnings.append(msg.format(s, i, wrn))

                else:
                    issues = []
                    fn = all_checks[mcfg[s][i].type]

                    b = fn(config=config_obj, item=i, section=s)
                    issues = b.check()

                    # Examine the issues
                    num_issues = len([True for p in issues if p != None])

                    for ii, issue in enumerate(issues):
                        if issue != None:
                            pi = i

                            # If we had a list, provide position
                            if num_issues > 1:
                                # Show 1 based lists
                                pi += "[{}]".format(ii + 1)

                            full_msg = msg.format(s, pi, issue)

                            if b.msg_level == 'warning':
                                warnings.append(full_msg)
                            else:
                                errors.append(full_msg)

    return warnings, errors


def cast_all_variables(config_obj, mcfg_obj):
    """
    Cast all values into the appropiate type using checkers, other_types
    and the master config.

    Args:
        config_obj: The object of the user config from
        mcfg_obj: The object used for manage the master config from
                  class MasterConfig
        other_types: User provided list to add any custom types


    Returns:
        ucfg: The users config dictionary containing the correct value types

    """

    ucfg = config_obj.cfg
    mcfg = mcfg_obj.cfg

    # Grab a dictionary of all the checkers
    all_checks = get_merged_checkers(config_obj)

    # Confirm checks are valid
    check_types(mcfg, all_checks)

    # Cast all variables
    for s in ucfg.keys():
        if s in mcfg.keys():
            for i in ucfg[s].keys():
                values = []

                # Ensure it is an item we can check
                if i in mcfg[s].keys():
                    fn = all_checks[mcfg[s][i].type]

                    b = fn(config=config_obj, section=s, item=i)

                    # cfg will be checked later all at once
                    values = b.cast()

                    # Developers can specify which items are lists in master cfg
                    if not mcfg[s][i].listed:
                        ucfg[s][i] = mk_lst(values, unlst=True)
                    else:
                        ucfg[s][i] = mk_lst(values)

                # Not recognized items, keep them anyways
                else:
                    values.append(ucfg[s][i])

    config_obj.cfg = ucfg
    return config_obj


def get_user_config(config_file, master_files=None, modules=None,
                    mcfg=None, changelog_file=None, cli=False):
    """
    Returns the users config as the object UserConfig.

    Args:
        config_file: real path to existing config file
        master_file: real path to a Core Config file
        modules: a module or list of modules with a string attribute
                __CoreConfig__ which is the path to a CoreConfig
        mcfg: the master config object after it has been read in.
        changelog_file: Path to a changlog showing any changes to the config
                        file the developers have made
        cli: boolean determining whether to attempt to process the changes
            which should be done from the CLI only

    Returns:
        ucfg: Users config as an object
    """

    if modules == None and master_files == None and mcfg == None:
        raise IOError("ERROR: Please provide either a module or a path to a"
                      " master config, or a master config object")
        sys.exit()

    if os.path.isfile(config_file):

        if master_files != None or modules != None:
            if master_files != None:
                master_files = mk_lst(master_files)
            if modules != None:
                modules = mk_lst(modules)

            mcfg = MasterConfig(path=master_files, modules=modules)

    else:
        raise IOError("Config file path {0} doesn't exist."
                      "".format(config_file))

    # Get users config object
    ucfg = UserConfig(config_file, mcfg=mcfg)

    # If were not running the CLI, raise exceptions for issues
    # Check out any change logs for issues
    chlog = ChangeLog(paths = ucfg.mcfg.changelogs, mcfg=ucfg.mcfg)
    potentials, required = chlog.get_active_changes(ucfg)

    # Required Changes that broke things
    if len(required) != 0 and not cli:
        cmd = get_inicheck_cmd(config_file, modules=modules,
                                            master_files=master_files)

        raise ValueError("\n\nUser's Config has deprecated information and"
                        " needs adjustment. To see what needs to change:"
                        "\n\n>> {}".format(cmd))

    # Fill in the gaps and make sure they're the right types
    ucfg.apply_recipes()
    ucfg = cast_all_variables(ucfg, mcfg)

    return ucfg

def config_documentation(out_f, paths=None, modules=None,
                        section_link_dict={}):
    """
    Auto documents the core config file. Outputs to a file which is can then be
    used for documentation. Specifically formulated for sphinx

    Args:
        out_f: string path to output location for the auto documentation
        paths: paths to master config files to use for creating docs
        modules: modules with attributes __core_config__ for creating docs
        section_link_dict: dictionary containing special documentation
                 for a section in the config file reference

    """

    if paths == None and modules == None:
        raise ValueError("inicheck function config_documentation args paths or"
                        " module must be specified!")


    master = MasterConfig(path=paths, modules=modules)
    mcfg = master.cfg

    # Beginning
    config_doc = "\nFor configuration file syntax information please visit"
    config_doc += " http://inicheck.readthedocs.io/en/latest/\n\n"

    for section in mcfg.keys():
        # Section header
        config_doc += "\n"
        config_doc += "{0}\n".format(section)
        config_doc += "-" * len(section) + '\n'

        # Allow for custom info in a description about the section, useful
        # for linking modules for auto documenting
        if section in section_link_dict.keys():
            intro = section_link_dict[section]
        else:
            intro = ''

        config_doc += intro
        config_doc += "\n"

        # Auto document config file according to master config contents
        for item,v in sorted(mcfg[section].items()):
            # Check for attributes that are lists
            for att in ['default', 'options']:
                z = getattr(v,att)
                if type(z) == list:
                    combo = ' '
                    doc_s = combo.join([str(s) for s in z])
                    setattr(v, att, doc_s)

            # Bold item with definition
            config_doc += "| **{0}**\n".format(item)

            # Add the item description
            config_doc += "| \t{0}\n".format(v.description)

            # Default
            config_doc += "| \t\t*Default: {0}*\n".format(v.default)

            # Add expected type
            config_doc += "| \t\t*Type: {0}*\n".format(v.type)

            # Print options should they be available
            if v.options:
                config_doc += "| \t\t*Options:*\n *{0}*\n".format(v.options)

            config_doc += "| \n"

            config_doc += "\n"

    path = os.path.abspath(out_f)
    if not os.path.isdir(os.path.dirname(path)):
        raise IOError("Config Documentation output file does not exist."
                      "\n{0}".format(out_f))

    print("Writing auto documentation for config file to:\n{0}".format(path))

    with open(path,'w+') as f:
        f.writelines(config_doc)
    f.close()



'''
Functions for checking values in a config file and producing errors and warnings
'''

import os
from pandas import to_datetime
from .utilities import mk_lst, is_valid, is_kw_matched, get_kw_match
import requests


class GenericCheck(object):
    """
    Generic Checking class. Every class thats a checker should inherit from this
    class. This class is used like:

    Every check will run the check().

    Attributes:
        message: String message to report if the value passed is not valid
        msg_level: Urgency of the message which can be either a warning or error
        values: value to be checked, casted, and reported on
        config: UserConfig object that the item/value being check resides
        type: string name representing the datatype which is based off the class
              name
        is_list: Boolean specifying the resulting values as a list or not
        allow_none: Boolean specifying the values can contain a value

        E.G.
            b = GenericCheck(section='test', item= 'tiem1', config=ucfg)
            issue = b.check()
            if issue != None:
                log.info(b.message)
    """

    def __init__(self, **kwargs):
        """
        Instatiates the check and setups the message, value and msg_level.

        Args:
            section: Name of the section contain the value being evaluated
            item: Name of the item in section which has a value being evaluated
            config: UserConfig object containing

        Raises:
            ValueError: Raises an error if Kwargs section, item, config is not
                        provided.
        """

        # Keywords are more convenient to use, make these ones required
        required = ["section", "item", "config"]
        for req in required:
            if req not in kwargs.keys():
                raise ValueError("Must provided at least keywords {} to "
                                "Checkers.".format(", ".join(required)))
            else:
                setattr(self,req,kwargs[req])

        # initial error messgae is nothing
        self.message = None

        # Initial level of concern is just a warning (can also be error)
        self.msg_level = 'warning'

        if not self.msg_level.lower() in ['warning', 'error']:
            raise ValueError("msg_level = {0} not allowed."
                             "".format(self.msg_level))

        # Initial values are set from the config directly, can be a list
        self.values = self.config.cfg[self.section][self.item]

        # Are the values received supposed to be a list?
        self.is_list = self.config.mcfg.cfg[self.section][self.item].listed

        # Allow None as a value?
        self.allow_none = self.config.mcfg.cfg[self.section][self.item].allow_none

        # Auto retrieve the type name from the class name which is always Check<type name>
        self.type = type(self).__name__.lower().replace('check','')


    def is_it_a_lst(self, values):
        """
        This checks to see if the original entry was a list or not.
        So instead of evaluating self.values  which is always a single item,
        we evaluate self.config[self.section][self.item]

        Args:
            values: The uncasted entry from a config file section and item,
                    can be a list or a single item

        Returns:
            boolean: True if its a list false if it is not.
        """
        # Grab the original values and see if theyre in a list
        if type(values) == list:
            return True

        # not a list, its good
        else:
            return False

    def is_valid(self, value):
        """
        Abstract function for defining how a value is checked for validity.
        It should always be used in check(). Is valid should always return
        a boolean whether the result is valid, and the issue as a string stating
        the problem. If no issue it should still return None.

        Args:
            value: Single value to be evaluated

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.
        """
        pass

    def check(self):
        """
        Function that is ran by the checking function of the user config to
        return useful message to instruct user.

        Returns:
            msgs: None if the entry is valid, else returns self.message
        """
        valids = []
        issues = []

        for v in mk_lst(self.values):
            valid, issue = self.is_valid(v)
            issues.append(issue)

        return issues


class CheckType(GenericCheck):
    """
    Base class for type checking whether a value of the right type. Here extra
    Attributes are added such as maximum and minimum.

    Attributes:
            minimum: Minimum value for bounded entries.
            maximum: Maximum value for bounded entries.
            type_func: Function used to cast the data to the desired type.
                      Default - str()
            bounded: Boolean indicating if a value can be limited by a min or max.
    """

    def __init__(self, **kwargs):
        super(CheckType, self).__init__(**kwargs)

        req = ['maximum', 'minimum']
        for kw in req:
            if kw in kwargs.keys():
                value = kwargs[kw]
            else:
                value = None

            setattr(self, kw, value)

        # Default Function used for casting to types
        self.type_func = str

        # Allow developers to specify bounds for certain types
        self.bounded = False

        # Default issue for type check is error
        self.msg_level = 'error'


    def check_bounds(self, value):
        """
        Checks the users values to see if its in the bounds specified by the
        developer.

        If self.max or self.min == None then it is assumed no bounds on either
        end.

        Function only runs when bounded = True in the master config.

        Args:
            value: Single value being evaluated against the bounds.

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.
        """

        valid = True
        msg = None

        if self.bounded:
            msg = "Value must be"

            # Grab it from the config
            if self.config != None:
                max_value = self.config.mcfg.cfg[self.section][self.item].max
                min_value = self.config.mcfg.cfg[self.section][self.item].min

            # Check upper and lower bounds
            if value != None:
                value = self.type_func(value)

                if min_value != None:
                    min_value = self.type_func(min_value)
                    msg += " greater than {}".format(min_value)

                    if value < min_value:
                        valid = False

                if max_value != None:
                    if min_value != None:
                        msg += " and"

                    max_value = self.type_func(max_value)
                    msg += " less than {}".format(max_value)

                    if value > max_value:
                        valid = False

            # Throw error if max or min is set and value is none.
            elif value == None and (max_value != None or min_value != None):
                valid == False
                msg = "Value cannot be None"

            if valid:
                msg = None

        return valid, msg

    def check_list(self):
        """
        Checks to see if self.values provided are in a list and if they should be.

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable in terms of being a list
                **msg** - string to print if value is not valid.
        """

        # Is it currently a list
        currently_a_list = self.is_it_a_lst(self.values)

        valid = True
        msg = None

        # NOTE This is for checking single items but we can cast these no matter what so its viable.
        # # Is it supposed to be a list and isn't?
        # if self.is_list and not currently_a_list:
        #     valid = False
        #     msg = "Expected list recieved single item"

        # Not supposed to be a list and is one?
        if not self.is_list and currently_a_list:
            msg = "Expected single value received list."
            valid = False

        return valid, msg

    def check_options(self, value):
        """
        Check to see if the current value being evaluated is also in the list of
        provided options in the master config. Only runs if options were
        provided.

        Args:
            value: A single value which is checked against the list of options
                   in the master config

        Returns:
            tuple:
                **valid** - Boolean whether the value was in the options list
                **msg** - string to print if value is not in the options.

        """
        valid = True
        msg = None

        if self.config.mcfg.cfg[self.section][self.item].options:

            # If it is not in the options its invalid
            options = self.config.mcfg.cfg[self.section][self.item].options

            if str(value).lower() not in options:
                msg = "Not a valid option"
                valid = False

        return valid, msg

    def is_valid(self, value):
        """
        Checks  a single value using the specified type function.
        All checkers should have a version of this function.

        Args:
            value: Single value to be evaluated

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.
        """
        valid, msg = is_valid(value, self.type_func, self.type,
                                             allow_none=self.allow_none)
        return valid, msg

    def check_none(self, value):
        """
        Check a single value if it is None and whether that is accecptable.

        Args:
            value: single value to be assessed whether none is valid

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.
        """
        # You got nones and you can't have them
        if not self.allow_none and str(value).lower() == 'none':
                valid = False
                msg = "Value cannot be None"
        else:
            valid = True
            msg = None

        return valid, msg


    def check(self):
        """
        Types are checked differently than some general checker.
        The checking goes through 5 steps and if anyone of them is invalid
        it will not check the rest. The process is as follows:

        1. Check if self.values should be a list or not.

        2. Check if a value in self.values should be None or not.

        3. Check for options and if a single value from self.values is among them.

        4. Check is a single value is valid according to self.valid.

        5. Check if a single value is inside any defined bounds.


        Returns:
            list: A list equal to the len(self.values) of either None or
                  strings relaying the issues found
        """
        msgs = []
        valids = []


        # 1. Check for lists
        valid, msg = self.check_list()
        msgs.append(msg)
        valids.append(valid)

        if valid:
            # We clear the nones since check_list is only checking the whole entry
            msgs = []
            valids = []

            for v in mk_lst(self.values):

                # 2. Check if none is allowed.
                valid, msg = self.check_none(v)

                if str(v).lower() != "none":
                    # 3. Check for option constraints
                    if valid:
                        valid, msg = self.check_options(v)

                    # 4. Check for type constraints
                    if valid:
                        valid, msg = self.is_valid(v)

                    # 5. Check for bounding constraints
                    if valid:
                        valid, msg = self.check_bounds(v)

                msgs.append(msg)
                valids.append(valid)

        return msgs

    def cast(self):
        """
        Attempts to return the casted values of the each value in self.values.

        This is performed with self.type_func unless the value is none in which
        we return None (NoneType)

        Returns:
            list: All values from self.values casted correctly
        """

        result = []

        for v in mk_lst(self.values):

            # 1. Manage nones:
            if str(v).lower() == "none":
                result.append(None)

            # 2. Manage the value types
            else:
                result.append(self.type_func(v))

        # 3. Manage the list
        if not self.is_list:
            result = mk_lst(result, unlst=True)

        return result

class CheckDatetime(CheckType):
    """
    Check values that are declared as type datetime. Parses anything that pandas
    to_datetime can parse.
    """

    def __init__(self, **kwargs):

        super(CheckDatetime, self).__init__(**kwargs)
        self.type_func = to_datetime


class CheckDatetimeOrderedPair(CheckDatetime):
    """
    Checks to see if start and stop based items are infact in fact ordered
    in the same section.

    Requires keywords section and item to ba passed as keyword args.
    Looks for keywords start/begin or stop/end in an item name. Then looks for a
    corresponding match with the opposite name.

    .. code-block:: ini

        start_simulation: 10-01-2019
        stop_simulation: 10-01-2017

    Would return an issue.

    This when checking the start will look for "simulation" with a
    temporal keyword reference in an item name like end/stop etc.
    Then it will attempt to determine them to be before.
    """

    def __init__(self, **kwargs):
        super(CheckDatetimeOrderedPair, self).__init__(**kwargs)

        self.cfg_dict = kwargs['config'].cfg[self.section]
        self.msg_level = "error"

    def is_corresponding_valid(self):
        """
        Looks in the config section for an opposite match for the item.

        e.g.
             if we are checking start_simulation, then we look for
             end_simulation

        Returns:
            corresponding: item name in the config section corresponding with
                           item being checked
        Raises:
            ValueError: raises an error if the name contains both sets or None
                        of keywords
        """
        init_kw = ["start",'begin']
        final_kw = ['stop','end']

        # First check whether our item has a kw
        is_start = is_kw_matched(self.item, init_kw)
        is_end = is_kw_matched(self.item, final_kw)

        # The current item is the beginning of an ordered datetime pair
        if is_start and not is_end:
            # Look for a corresponding end
            corresponding = get_kw_match(self.cfg_dict.keys(), final_kw)

        # The current item is the end of an ordered datetime pair
        elif is_end and not is_start:
            # Look for the corresponding start
            corresponding = get_kw_match(self.cfg_dict.keys(), init_kw)

        else:
            raise ValueError("Ordered Datetime pairs must be distinguishable "
                             " by item name. {} was either found to have both "
                             " sets of keywords or none of them."
                             "".format(self.item))

        # Is corresponding castable?
        corresponding_val = self.cfg_dict[corresponding]
        valid, msg = is_valid(corresponding_val, self.type_func, self.type)

        if valid:
            corresponding_val = self.type_func(corresponding_val)
            value = self.type_func(self.values )

            if is_start:
                # validity check
                order_valid = value < corresponding_val
                incorrect_context = "after"

            elif is_end:
                # Check if the
                order_valid = value > corresponding_val
                incorrect_context = "before"

            msg = "Date is {} {} value".format(incorrect_context, corresponding)

        valid = valid and order_valid

        return valid, msg

    def is_valid(self, value):
        """
        Checks whether it convertable to datetime, then checks for order.

        Args:
            value: Single value to be evaluated

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.
        """
        valid, msg = is_valid(value , to_datetime, self.type)

        if valid:
            valid, msg = self.is_corresponding_valid()

        return valid, msg


class CheckFloat(CheckType):
    """
    Float checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckFloat, self).__init__(**kwargs)
        self.type_func = float

        # Can be bounded but not required
        self.bounded = True

class CheckInt(CheckType):
    """
    Integer checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckInt, self).__init__(**kwargs)
        self.type_func = self.convert_to_int

        # Can be bounded but not requried
        self.bounded = True

    def convert_to_int(self, value):
        """
        When expecting an integer, it is convenient to automatically convert
        floats to integers (e.g. 6.0 --> 6) but its pertinent to catch when the
        input has a non-zero decimal and warn user (e.g. avoid 6.5 --> 6)

        Args:
            value: The value to be casted to integer
        Returns:
            self.values : the value converted
        """

        self.values  = float(value)

        if self.values .is_integer():
            self.values  = int(self.values)

        else:
            raise ValueError("Expecting integer and received float with "
                            " non-zero decimal")
        return self.values


class CheckBool(CheckType):
    """
    Boolean checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckBool, self).__init__(**kwargs)
        self.type_func = self.convert_bool

        self.affirmatives = ['y', 'yes', 'true']
        self.negatives = ['n', 'no', 'false']

    def convert_bool(self, value):
        """
        Function used to cast values to boolean

        Args:
            value:  value(s) to be casted

        Returns:
            value: Value returned as a boolean
        """

        v = str(value).lower()

        if v in self.affirmatives:
            result = True

        elif v in self.negatives:
            result = False

        else:
            raise ValueError("Value {0} not coercable to boolean."
                             "".format(value))

        return result


class CheckString(CheckType):
    """
    Float checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckString, self).__init__(**kwargs)

        # Most strings types that are not paths should be lower case
        self.type_func = lambda x: str(x).lower()


class CheckPath(CheckType):
    """
    Checks whether a Path exists. Base for checking if paths exist.
    """

    def __init__(self, **kwargs):

        super(CheckPath, self).__init__(**kwargs)

        # Allow None as a value?
        self.allow_none = False

        self.root_loc = os.path.dirname(self.config.filename)
        self.dir_path = False
        self.type_func = self.make_abs_from_cfg

    def is_valid(self, value):
        """
        Checks for existing filename

        Args:
            value: Single value to be evaluated

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.
        """

        v = os.path.join(self.root_loc, value)
        if self.dir_path:
            valid = os.path.isdir(v)

        else:
            valid = os.path.isfile(v)

        if valid:
            msg = None
        else:
            msg = self.message

        return valid, msg

    def make_abs_from_cfg(self, value):
        """
        Looks at a path and determines its absolute path. All paths should be
        either absolute or relative to the config file in which they will be
        converted to absolute paths

        Args:
            value: single path or filename
        Returns:
            str: absolute path or filename
        """
        if not os.path.abspath(value):
            value = os.path.join(self.root_loc, value)

        return value


class CheckDirectory(CheckPath):
    """
    Checks whether a directory exists. These directories are allowed to be none
    and only warn when they do not exist. E.g. Output folders
    """

    def __init__(self, **kwargs):
        super(CheckDirectory, self).__init__(**kwargs)
        self.dir_path = True
        self.allow_none = True
        self.message = "Directory does not exist."

class CheckFilename(CheckPath):
    """
    Checks whether a directory exists. These are files that the may be created
    or not necessary to run. E.g. Log files
    """

    def __init__(self, **kwargs):
        super(CheckFilename, self).__init__(**kwargs)
        self.message = "File does not exist."
        self.allow_none = True


class CheckCriticalFilename(CheckFilename):
    """
    Checks whether a critical file exists. This would be any files that
    absolutely have to exist to avoid crashing the software. These are static
    files.
    """

    def __init__(self, **kwargs):
        super(CheckCriticalFilename, self).__init__(**kwargs)
        self.msg_level = 'error'
        self.allow_none = False


class CheckDiscretionaryCriticalFilename(CheckCriticalFilename):
    """
    Checks whether a an optional file exists that may change software behavior
    by being present. In other words, this can be none and still valid. If the
    not then it registers an error if the string doesn't exist as path.
    """

    def __init__(self, **kwargs):
        super(CheckDiscretionaryCriticalFilename, self).__init__(**kwargs)
        self.allow_none = True


class CheckCriticalDirectory(CheckDirectory):
    """
    Checks whether a critical directory exists. This is for any directories
    that have to exist prior to launching some piece of software.
    """

    def __init__(self, **kwargs):
        super(CheckCriticalDirectory, self).__init__(**kwargs)
        self.msg_level = 'error'
        self.allow_none = True

class CheckURL(CheckType):
    """
    Check URLs to see if it can be connected to.
    """

    def __init__(self, **kwargs):

        super(CheckURL, self).__init__(**kwargs)
        self.msg_level = 'error'

    def is_valid(self, value):
        """
        Makes a request to the URL to determine the validity.

        Args:
            value: Single value to be evaluated

        Returns:
            tuple:
                **valid** - Boolean whether the value was acceptable
                **msg** - string to print if value is not valid.

        """
        # Attempt to establish a connection
        try:
            r = requests.get(value, timeout=5)

        except Exception as e:
            msg = "Invalid connection or URL"
            r = None


        valid = False
        msg = "Webpage does not exist"

        if r != None:
            if r.status_code == 200:
                valid = True
                msg = None

        return valid, msg



import sys
import os


def find_options_in_recipes(recipes, choice_search, action_kw, condition_position=0):
    """
    Looks through the master config at recipes and entries to determine if
    there are place the developer made distince choices, this is used in the
    inimake script to walk users through making their own config file from
    scratch.

    This function looks through and finds sections or items that match the
    conditional. Then it looks through the adj_config to determine if any
    sections/items are being removed as a result, then it finds those.
    If a section/item appears in a recipe condition and a removed by
    another trigger it is extremely likely these are choices.

    Args:
        mcfg: Recipes object from master config
        conditional: list of 3 focusing on what were looking for
        action_kw: action keyword is a string of what to look for in matching
                   choices where ever true is used will popoulate with every
                   item/section

    Returns:
        final_choices:

    """
    decisions = {}
    conditional = ["any" for i in range(3)]
    final_choices = []
    pos = condition_position

    # Gather which recipes might represent a decision point
    for r in recipes:
        choices = []

        for t_name, trigger in r.triggers.items():

            for condition in trigger.conditions:
                # if the position of interest is not a generic place holder
                if condition[pos] != "any":
                    # Look at how it adjusts the config
                    for k, v in r.adj_config.items():
                        # If our action word shows up then we know its a choice
                        if action_kw in v.keys():
                            choices.append(condition[pos])
                            # Look inside the adjustment for more choices
                            for kk,vv in v.items():
                                actions = [k,kk,vv]

                                if kk == action_kw:
                                    choices.append(actions[pos])

        # if the list is not empty add it

        if choices:
            final_choices.append(tuple(set(sorted(choices))))

    # Because there will be redundent sets of choices, we perform a set to clean up
    return list(set(final_choices))


def ask_user(question, choices=None):
    """
    Asks the user which choice to make and handles incorrect choices.
    """
    valid = False
    if choices == None:
        choices = ["yes", "no"]

    if choices != None:
        choice_str = "(" + ", ".join(choices) + ")\n"

    while not valid:
        msg = question + choice_str
        s = input(msg)
        if s not in choices:
            print("Invalid Choice: Try again.")
            valid = False
        else:
            valid = True
    return s


def ask_config_setup(choices_series, section=None, item=None,
                                                   num_questions=None):
    """
    Ask repeating questions for setting up a config
    """
    if section == None and item == None:
        msg =  "Which of these sections do you want to use? "

    elif section != None and item == None:
        msg =  ("In section {}, Which of these items do you want to use? "
               "".format(section))

    else:
        msg =  ("In section {} item {},  Which of these values do you want to "
               "use? ".format(section, item))

    if num_questions == None:
        num_questions = len(choices_series)

    # Add sections that we choose
    selections = []

    for i, c in enumerate(choices_series):
        count = "{}/{}: ".format(i + 1, num_questions)
        q = count + msg

        selections.append(ask_user(q, c))

    return selections


def remove_comment(string_entry):
    """
    Takes a single line and removes and .ini type comments.
    Also remove any spaces at the begining or end of a commented line
    """
    # Comment checking
    if '#' in string_entry:
        result = string_entry.split('#')[0]
    elif ';' in string_entry:
        result = string_entry.split(';')[0]
    else:
        result = string_entry

    return result


def mk_lst(values, unlst=False):
    """
    while iterating through several type of lists and items it is convenient to
    assume that we recieve a list, use this function to accomplish this. It also
    convenient to be able to return the original type after were done with it.
    """
    # Not a list, were not looking to unlist it. make it a list
    if type(values) != list and not unlst:
            values = [values]

    # We want to unlist a list we get
    if unlst:
        # single item provided so we don't want it a list
        if type(values) == list:
            if len(values) == 1:
                values = values[0]

    return values


def remove_chars(orig_str, char_str, replace_str=None):
    """
    Removes all character in orig_str that are in char_str

    Args:
        orig_str: Original string needing cleaning
        char_str: String of characters to be removed
        replace_str: replace values with a character if requested

    Returns:
        string: orig_str with out any characters in char_str
    """
    if replace_str == None:
        clean_str = [c for c in orig_str if c not in char_str]
    else:
        clean_str = [c if c not in char_str else replace_str for c in orig_str]

    return "".join(clean_str)


def get_relative_to_cfg(path, user_cfg_path):
    """
    Converts a path so that all paths are relative to the config file

    Args:
        path: relative str path to be converted
        user_cfg_path: path to the users config file
    Returns:
        path: absolute path of the input considering it was relative to the cfg
    """
    if not os.path.isabs(path):
        path = os.path.abspath(os.path.join(os.path.dirname(user_cfg_path),
                                                            path))
    return path


def is_kw_matched(single_word, kw_list, kw_count=1):
    """
    Checks to see if there are any keywords in the single word and returns
    a boolean

    Args:
        single_word: String to be examined for keywords
        kw_list: List of strings to look for inside the single word
        kw_count: Minimum Number of keywords to be found for it to be True

    Returns:
        boolean: Whether a keyword match was found
    """
    truths = [True for kw in kw_list if kw in single_word]
    if len(truths) >= kw_count:
        return True
    else:
        return False


def get_kw_match(potential_matches, kw_list, kw_count=1):
    """
    Loops through a list of potential matches looking for keywords in the
    list of strings being presented. When a match is found return its value.

    Args:
        potential_matches: List of strings to check
        kw_list: List of strings to look for inside the single word
        kw_count: Minimum Number of keywords to be found for it to be True

    Returns:
        result: The first potential found with the keyword match
    """
    result = False

    for potential in potential_matches:
        match = is_kw_matched(potential, kw_list, kw_count=kw_count)

        if match:
            result = potential
            break

    return result


def pcfg(cfg):
    """
    prints out the config file to the prompt with a nice stagger Look for
    readability

    Args:
        cfg: dict of dict in a heirarcy of  {section, {items, values}}
    """

    for sec in cfg.keys():
        print(repr(sec))
        try:
            for item in cfg[sec].keys():
                print('\t' + item)
                values = cfg[sec][item]

                if type(cfg[sec][item]) == list:
                    out = ", ".join(cfg[sec][item])
                else:
                    out = cfg[sec][item]
                print('\t\t' + repr(out))
        except:
            print('\t recipe')


def pmcfg(cfg):
    """
    prints out the core config file to the prompt with a nice stagger Look for
    readability

    Args:
        cfg: dict of dict in a hierarchy of  {section, {items, values}}
    """

    for sec in cfg.keys():
        print(repr(sec))
        for item in cfg[sec].keys():
            print('\t' + item)
            obj = cfg[sec][item]
            for att in ['type', 'options', 'default', 'description']:
                value = getattr(obj, att)
                print('\t\t' + att)

                if type(value) == list:
                    out = ", ".join(value)
                else:
                    out = value
                print('\t\t\t' + repr(out))

def is_valid(value, cast_fn, expected_data_type, allow_none=False):
    """
    Checks whether a value can be converted using the cast_fn function.

    Args:
        value: Value to be considered
        cast_fn: Function used to determine the validity, should throw an
                 exception if it cannot
        expected_data_type: string name of the expected data
        allow_none: Boolean determining if none is valid
    Returns:
        tuple:
            **valid (boolean)**: whether it could be casted
            **msg (string)**: Msg reporting what happen
    """
    try:

        if value != None:
            value = cast_fn(value)
            valid = True
            msg = None

        elif allow_none and value == None:
            valid = True
            msg = None

        else:
            valid = False


    except Exception as e:
        valid = False
        msg = "Expecting {0} received {1}".format(expected_data_type,
                                                  type(value).__name__)
    return valid, msg

def get_inicheck_cmd(config_file, modules=None, master_files=None):
    """
    Strings together an inicheck cli command based on modules and files

    """

    cmd = "inicheck -f {}".format(config_file)

    if master_files != None:
        master_files = mk_lst(master_files)
        cmd += " -mf {}".format(" ".join(master_files))

    if modules != None:
        modules = mk_lst(modules)
        cmd += " -m {}".format(" ".join(modules))

    return cmd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_checkers
----------------------------------

Tests for `inicheck.checkers` module.
"""

import unittest
from inicheck.checkers import *
from inicheck.config import UserConfig, MasterConfig
import inicheck

class TestCheckers(unittest.TestCase):

    def run_a_checker(self, valids, invalids, checker, section='basic',
                                                       item='item',
                                                       extra_config={}):
        """
        Runs a loop over all the valids and applies the checker and asserts
        theyre true. Same thing is done for the invalids
        Args:
            valids: List of valid entries to check
            invalids: List of invalid entries to check
            checker: any class in inicheck.checkers
            ucfg: inicheck.config.UserConfig instance (optional)
            is_list: is it expected to be a list?
            section: section name the item being checked is occurring
            item: Item name in the config
            extra_config: Pass in contextual config info to test more complicated checkers. E.g. ordered datetime pair.
        """

        cfg = self.ucfg.cfg
        cfg.update({section:{item:" "}})

        # Added info for testing e.g. ordered datetime pair
        if extra_config:
            cfg.update(extra_config)

        for z,values in enumerate([valids, invalids]):
            for v in values:

                cfg[section][item] = v
                b = checker(config=self.ucfg, section=section, item=item)
                msgs = b.check()

                if len([True for m in msgs if m==None]) == len(msgs):
                    valid = True
                else:
                    valid = False

                # Expected valid
                if z == 0:
                    assert valid
                else:
                    assert not valid

    @classmethod
    def setUpClass(self):
        """
        Create some key structures for testing
        """
        tests_p = os.path.join(os.path.dirname(inicheck.__file__),'../tests')
        self.mcfg = MasterConfig(path=os.path.join(tests_p,
                                    'test_configs/master.ini'))

        self.ucfg = UserConfig(os.path.join(tests_p,"test_configs/base_cfg.ini"),
                               mcfg=self.mcfg)


    def test_string(self):
        """
        Test we see strings as strings
        """

        # Confirm we these values are valid
        valids = ['test']
        self.run_a_checker(valids, [], CheckString,  item='username')

        # Confirm that casting a string with uppers will auto produce lowers
        self.ucfg.cfg['basic']['username'] = 'Test'

        b = CheckString(config=self.ucfg, section='basic', item='username')
        result = b.cast()
        assert result == 'test'

    def test_bool(self):
        """
        Test we see booleans as booleans
        """

        # Confirm we these values are valid
        valids = [True, False, 'true', 'FALSE', 'yes', 'y', 'no', 'n']
        invalids = ['Fasle', 'treu']
        self.run_a_checker(valids, invalids, CheckBool, item='debug')


    def test_float(self):
        """
        Test we see floats as floats
        """
        valids = [-1.5, '2.5']
        invalids =  ['tough']

        self.run_a_checker(valids, invalids, CheckFloat, item='time_out')


    def test_int(self):
        """
        Test we see int as ints and not floats
        """

        # Confirm we these values are valid
        valids = [10, '2', 1.0]
        invalids = ['tough', '1.5', '']
        self.run_a_checker(valids, invalids, CheckInt, item='num_users')


    def test_datetime(self):
        """
        Test we see datetime as datetime
        """

        valids = ['2018-01-10 10:10','10-10-2018', "October 10 2018"]
        invalids = ['Not-a-date', 'Wednesday 5th']
        self.run_a_checker(valids, invalids, CheckDatetime, item='start_date')

    def test_list(self):
        """
        Test our listing methods using lists of dates.
        """

        valids = ['10-10-2019',['10-10-2019'],['10-10-2019','11-10-2019']]
        self.run_a_checker(valids, [], CheckDatetime, item='epochs')

        # # We have a list in the config when we don't want one
        # self.ucfg.cfg['section']['item'] = ["test", "test2"]
        # invalids = ['test']
        # self.run_a_checker([], invalids, CheckFilename,  is_list=False)

    def test_directory(self):
        """
        Tests the base class for path based checkers
        """

        valids = ["./"]
        invalids = ['./somecrazy_location!/']
        self.run_a_checker(valids, invalids, CheckDirectory, item='tmp')

    def test_filename(self):
        """
        Tests the base class for path based checkers
        """
        # Remember paths are relative to the config
        valids = ["../test_checkers.py"]
        invalids = ['dumbfilename']
        self.run_a_checker(valids, invalids, CheckFilename, item='log')

    def test_url(self):
        """
        Test our url checking.
        """
        valids = ["https://google.com"]
        invalids = ["https://micah_subnaught_is_awesome.com"]
        self.run_a_checker(valids, invalids, CheckURL,
                                                 item='favorite_web_site')

    def test_datetime_ordered_pairs(self):
        """
        Tests the ordered datetime pair checker which looks for <keyword>_start
        <keyword>_end pairs and confirms they occurs in the correct order.

        """
        valids = ["1-02-2019"]
        invalids = ["01-01-2018"]
        acfg = {'basic':{'start_date':'1-1-2019'}}
        self.run_a_checker(valids, invalids, CheckDatetimeOrderedPair,
                                             item="end_date",
                                             extra_config=acfg)
    def test_bounds(self):
        """
        MasterConfig options now have max and min values to constrain continuous
        types. This tests whether that works
        """

        self.run_a_checker([1.0, 0.0, '0.5'], [1.1,-1.0, '10'], CheckFloat,
                                                                item='fraction')

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_changes
----------------------------------

Tests for `inicheck.changes` module.
"""

import unittest
from inicheck.changes import *
from inicheck.config import UserConfig, MasterConfig
import os

class TestChanges(unittest.TestCase):

    def test_valid_syntax(self):
        """
        Test we can detect valid syntax
        """
        base = os.path.dirname(__file__)
        master = os.path.join(base,'test_configs/CoreConfig.ini')
        recipes = os.path.join(base,'test_configs/recipes.ini')
        mcfg = MasterConfig(path=[master, recipes])
        cf = os.path.join(base,'test_configs/changelog.ini')
        try:
            c = ChangeLog(paths=cf, mcfg=mcfg)
            assert True
        except Exception as e:
            assert False

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())
